#!/bin/bash

#######################################################################
#  Proprietary Information D-Wave Systems Inc.                        #
#  Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
#  Notice dw-convert is licensed to authorized users only under the   #
#  applicable license agreement.  See eula.txt                        #
#  D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
#######################################################################

# This utility scans through the user's workspaces and looks for files
# with suffix epqmi, qmi or sol.  For those files, a format conversion
# is applied.  The original files are copied to the corresponding
# location in $DWAVE_WORKSPACE/fmt2.  At the same time, we're going to
# change the name of dot files to eliminate the leading dot.  This
# includes these files:
#
#    .configuration*
#    .geometry
#    .epqmi
#    .qubo

# Make sure we get the correct "find" command on MinGW:

name=$(uname)

if [[ $name == *MINGW* ]]
then
    lspath=$(which ls)
    find=${lspath/ls/find}
else
    find=find
fi

# We are supporting multi-user mode, so here is where the workspaces
# are located:

if [[ -z $DWAVE_WORKSPACE ]]
then
    echo "Please make sure that your DWAVE_WORKSPACE environment variable is set"
    exit 1
fi

cd $DWAVE_WORKSPACE

# Create the list of workspaces to process:

for workspace in workspace.*
do
    workspace_list="${workspace_list} $workspace"
done

# Make corresponding directories inside $DWAVE_WORKSPACE/fmt2:

for workspace in ${workspace_list}
do
    echo "**************** Directory creation for $workspace ****************"

    dirlist=$($find $workspace -type d)

    for dir in $dirlist
    do
	mkdir -p $DWAVE_WORKSPACE/fmt2/$dir
    done
done

# Iterate over workspaces.  For each, copy the format 2 files into
# $DWAVE_WORKSPACE/fmt2 and then upconvert the individual files:

for workspace in ${workspace_list}
do
    echo "**************** Handling $workspace ****************"

    # Remove the dot in the name of the dot files:

    for dot_file in $workspace/.conf* $workspace/.geo*
    do
	[[ -e ${dot_file} ]] || continue
	fname=${dot_file#${workspace}/}
	fname_no_dot=${fname#.}
	echo "Removing dot from name in: ${fname}"
	mv $workspace/$fname $workspace/$fname_no_dot
    done

    if [[ -e $workspace/configuration1 ]]
    then
	conf_file=$workspace/configuration1
    else
	echo "No configuration1 file found in $workspace"
	exit 1
    fi

    # Parse the configuration1 file to get connection and solver:

    oldIFS="${IFS}"
    IFS="|"

    while read token1 token2
    do
	[[ $token1 == connection ]] && connection=$token2
	[[ $token1 == solver ]] && solver=$token2
    done < $conf_file

    IFS="${oldIFS}"

    # Configure for the workspace:

    dw set connection $connection
    dw set solver $solver

    # List the dot files that need to be renamed:

    file_list=$($find $workspace -name .qubo -or -name .epqmi)

    for file in $file_list
    do
	file=${file#${workspace}}
	file_name=${file##*/}
	wspath=${file%${file_name}}

	dw cd ${wspath}

	echo "Moving ${wspath}${file_name} to ${wspath}default${file_name}"
	dw mv ${file_name} default${file_name}
    done

    # List the files that may need format conversion and update them:

    file_list=$($find $workspace -name \*.qmi -or -name \*.sol -or -name \*.epqmi)

    for file in $file_list
    do
	file=${file#${workspace}}
	file_name=${file##*/}
	wspath=${file%${file_name}}

	dw cd ${wspath}

	format=$(dw-get-format ${file_name})

	status=$?
	if [[ $status != 0 ]]
	then
	    echo "Aborting conversion: error while processing ${wspath}${file_name}"
	    exit 1
	fi

	if [[ $format == 3 ]]
	then
	    echo "Skipping conversion (format == 3): ${wspath}${file_name}"
	    continue
	fi

	echo "Converting: ${wspath}${file_name}"

	path=$(dw get path)
	path=${path#*=}
	path=${path#\'}
	path=${path%\'}
	path=${path%/}
	fmt2path=${path/${DWAVE_WORKSPACE}/${DWAVE_WORKSPACE}\/fmt2}

	cp ${path}/${file_name} ${fmt2path}/${file_name}

	[[ ${file_name} == *.qmi ]] && dw get ${file_name} | dw set CONVERT.${file_name}
	[[ ${file_name} == *.sol ]] && dw-convert-sol ${file_name} CONVERT.${file_name}
	[[ ${file_name} == *.epqmi ]] && dw-convert-epqmi ${file_name} CONVERT.${file_name}

	dw mv CONVERT.${file_name} ${file_name}
    done

done
