#!/usr/bin/env python
#
# Proprietary Information D-Wave Systems Inc.
# Copyright (c) 2015 D-Wave Systems Inc. All rights reserved.
# Notice qbqual.py is licensed to authorized users only under the
# applicable license agreement.  See eula.txt.
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC V5G 4M9, Canada.
#
#
import sys
import os
import math
import numpy as np
import networkx as nx
import networkx.utils as nxu

def size(qubo):
    if len(qubo) == 0:
        ret = 0
    else:
        ret = max([it[0] if it[0] > it[1] else it[1] for it in qubo.keys()]) + 1
    return ret

def nLevels(quboSize):
    ATOMIC_SIZE = 8
    ret = int(math.ceil(math.log(float(quboSize)/ATOMIC_SIZE,2)))+1
    return ret

def extractSubqubo(qubo,verts,verbose=0):
    mapp = zip(verts,range(len(verts)))
    b2s = {}; s2b = {}      # big2small/small2big permutation
    for it in mapp:
        b2s[it[0]] = it[1]
        s2b[it[1]] = it[0]
    subG = {}
    for it in qubo.items():
        if it[0][0] in b2s and it[0][1] in b2s:
            subG[b2s[it[0][0]],b2s[it[0][1]]] = it[1]
    # ensure that all Vs are represented in the subQUBO
    reprVerts = set([it[0] for it in subG.keys()]+[it[1] for it in subG.keys()])
    if len(verts) != len(reprVerts):
        for i in range(len(verts)):
            if i in reprVerts:
                continue
            subG[i,i] = 0.0
            if verbose > 0:
                print("extractSubqubo: added subG["+str(i)+","+str(i)+"], verts="+str(verts)+", non-repr vert="+str(verts[i]))

    return (subG,b2s,s2b)

def readDIMACSQuboFile(fName, annotStr=None):
    inFile = open(fName,'r')
    # eliminate any initial comments
    try:
        line = inFile.next()
    except StopIteration:
        raise RuntimeError("empty QUBO file '"+fName+"'")
    while line[0]=='c':
        try:
            line = inFile.next()
        except StopIteration:
            raise RuntimeError("premature end to QUBO file '"+fName+"'")
    # header line should be next
    hdr = [item.strip() for item in line.split() if len(item)>0]
    if hdr[0]=='p':
        if hdr[1]=='qubo':
            if hdr[2]=='unconstrained' or int(hdr[2])==0:
                target = 'unconstrained'
            else:
                #print "unsupported target type '"+str(hdr[1])+"', exiting..."
                #sys.exit(-1)
                raise RuntimeError("unsupported target type '"+str(hdr[1])+"'")
            maxDiagonals = int(hdr[3])
            nDiagonals = int(hdr[4])
            nElements = int(hdr[5])
        else:
            #print "unsupported file type '"+line+"', exiting..."
            #sys.exit(-1)
            raise RuntimeError("unsupported file type '"+line+"'")
    spMat = {}
    backmap = {}
    els = 0
    # process the rest of the file.  If annotStr specified, digest those annotations properly
    #  and attach to the following line of the file.
    for line in inFile:
        adj = [item.strip() for item in line.split() if len(item)>0]
        if adj[0].startswith('c') and (not adj[0].startswith('c'+annotStr) if annotStr!=None else True):
            continue
        if annotStr and adj[0].startswith('c'+annotStr):
            annot = adj[1:]
            continue
        if not adj[0][0].isalnum():
            raise RuntimeError("first field in line is not numeric")
        try:
            floatAdj2 = float(adj[2])
        except:
            raise RuntimeError("third field in line is not float")
        if (int(adj[0]),int(adj[1])) in spMat:
            print "Warning: element ("+adj[0]+","+adj[1]+") occurs multiple times"
        els += 1
        spMat[int(adj[0]),int(adj[1])] = floatAdj2
        if annotStr:
            # if they exist, annot[3:] are any transform-specific values to be passed through
            backmap[int(adj[0]),int(adj[1])] = (int(annot[0]),int(annot[1]),float(annot[2]),annot[3:])
        annot = None
        #if verbose>2: print "els="+str(els)+" "+adj[0]+" "+adj[1]+(str(annot) if annotStr else "")

    # do some sanity-checking on file contents
    if els==0:
        return None
    #if len(spMat) != els and els != nDiagonals+nElements:
    #    print "Warning: duplicate elements in file: #elements="+str(len(spMat))+", #nondistinct elements="+str(els)
    nOffDiags = len([(it[0]) for it in spMat if it[0]!=it[1]])
    if nOffDiags != nElements:
        print "Warning: 'nElements' ("+str(nElements)+") does not match actual number ("+str(nOffDiags)+")"
    maxNdx =  max([max(it[0],it[1]) for it in spMat]) 
    if maxNdx >= maxDiagonals:
        print "Warning: 'maxDiagonals' ("+str(maxDiagonals)+") less than largest index ("+str(maxNdx)+")"
    nDiags = len([(it[0]) for it in spMat if it[0]==it[1]])
    if nDiags != nDiagonals:
        print "Warning: 'nDiagonals' ("+str(nDiagonals)+") does not match actual number ("+str(nDiags)+")"
    ret = {'spMat':spMat, 'target':target, 'maxDiagonals':maxDiagonals, 'nDiagonals':nDiagonals, 'nElements':nElements}
    if annotStr:
        ret['annotations'] = backmap
    return ret

# utility routine to read a raw QUBO file of the format
#        30 412
#        1 1 0
#        2 2 0
#        3 3 -1
#    ...
# where the first line is #variables, #edges
# and the remaining #edges lines are vertex1, vertex, and coupler strength

def readRawQuboFile(args, fName):
    inFile = open(fName,'r')
    # header line should be first
    line = inFile.next()
    hdr = [item.strip() for item in line.split() if len(item)>0]
    if not hdr[0].isdigit() or not hdr[1].isdigit():
        print "header line has non-integer number of variables (" +str(hdr[0])+") or number of edges ("+str(hdr[1])+"), exiting..."
        sys.exit(-1)
    nDiagonals = int(hdr[0])
    nElements = int(hdr[1])
    spMat = {}
    els = 0
    # process the rest of the file as adjacency entries
    for line in inFile:
        adj = [item.strip() for item in line.split() if len(item)>0]
        els += 1
        spMat[int(adj[0]),int(adj[1])] = float(adj[2])
        if args.verbose>1: print "els="+str(els)+" "+adj[0]+" "+adj[1]

    # do some sanity-checking on file contents
    if args.verbose>0:
        print "done reading QUBO file, #elements="+str(len(spMat))
    if els != nDiagonals+nElements:
        print "Warning: duplicate elements in file: #elements="+str(len(spMat))+", #nondistinct elements="+str(els)
    nOffDiags = len([(it[0]) for it in spMat if it[0]!=it[1]])
    if nOffDiags != nElements:
        print "Warning: 'nElements' ("+str(nElements)+") does not match actual number ("+str(nOffDiags)+")"
    maxNdx =  max([max(it[0],it[1]) for it in spMat]) 
    if maxNdx >= nDiagonals:
        print "Warning: 'nDiagonals' ("+str(nDiagonals)+") less than largest index ("+str(maxNdx)+")"
    nDiags = len([(it[0]) for it in spMat if it[0]==it[1]])
    if nDiags != nDiagonals:
        print "Warning: 'nDiagonals' ("+str(nDiagonals)+") does not match actual number ("+str(nDiags)+")"
    ret = {'spMat':spMat, 'nDiagonals':nDiagonals, 'nElements':nElements}
    return ret

def readQuboFile(fName):
    ch = open(fName).read(1)
    if ch=='':
        ret = None    # null file
    elif ch=='c' or ch=='p':
        ret = readDIMACSQuboFile(fName)
    else:
        ret = readRawQuboFile(fName)
    return ret
    
def writeDIMACSQuboFile(args, h, J, fName):
    wFile = open(fName,'w')
    #quboSz = max
    if len(J.items()) > 0:
        quboSz = 1 + max([it[0][0] if it[0][0] > it[0][1] else it[0][1] for it in J.items()])
    else:
        print "Warning: no J elements specified; assuming QUBO size is the length of h"
        quboSz = len(h)
    #if quboSz > len(h):
    #    print "Warning: size of 'h' ("+str(len(h))+") less than maximum value in 'J' ("+str(quboSz)+")"
    if quboSz < len(h):
        quboSz = len(h)
    nDiagonals = len([it for it in h if it])
    wFile.write("p qubo unconstrained "+str(len(h))+" "+str(nDiagonals)+" "+str(len(J))+"\n")
    wFile.write("c   QUBO file written by qforge.py \n")
    wFile.write("c   First, the 'h' values --------------------\n")
    for (ndx,val) in zip(range(len(h)),h):
        if not val:
            continue
        wFile.write(str(ndx)+" "+str(ndx)+" "+str(val)+"\n")
    wFile.write("c   Then the 'J' values --------------------\n")
    for edge in sorted(J.items()):
        wFile.write(str(edge[0][0])+" "+str(edge[0][1])+" "+str(edge[1])+"\n")
    wFile.flush()
    wFile.close()
    return

#   writeAnnotDIMACSQuboFile: write an annotated DIMACS-style QUBO file, where the 
#       annotations are embodied by call-back functions, so can be used to insert
#       "symbol table"-style info about which variables/links a transformed QUBO's
#       variables/links came from.
#   Inputs:
#   - args: from the command line, as from argparse
#   - h: 
#   - J: 

def writeAnnotDIMACSQuboFile(args, h, J, fName, hFn, hArgs, jFn, jArgs):
    wFile = open(fName,'w')
    maxHVar = max([it[0][0] for it in h])
    maxJVar = max([it[0][0] if it[0][0] > it[0][1] else it[0][1] for it in J]) 
    quboSz = 1 + max(maxHVar,maxJVar)
    wFile.write("p qubo unconstrained "+str(quboSz)+" "+str(len(h))+" "+str(len(J))+"\n")
    wFile.write("c Legend for line above:\n")
    wFile.write("c   <program marker> <type> <subtype> <maxDiagonals> <nDiagonals> <nOffdiagonals>\n")
    wFile.write("c   QUBO file written by qforge.py \n")
    wFile.write("c Legend for non-comment lines following:\n")
    wFile.write("c <var1> <var2> <weight>\n")
    wFile.write("c   First, the 'h' values --------------------\n")
    for ((ndx,val),hArg) in zip(h,hArgs):
        hFn(wFile,(ndx,val),hArg)
        wFile.write(str(ndx[0])+" "+str(ndx[0])+" "+str(val)+"\n")
    wFile.write("c   Then the 'J' values --------------------\n")
    for (edge,jArg) in zip(J,jArgs):
        jFn(wFile,edge,jArg)
        wFile.write(str(edge[0][0])+" "+str(edge[0][1])+" "+str(edge[1])+"\n")
    wFile.flush()
    wFile.close()
    return

def writeSolutionFile(args, solutions, fName): #, hFn, hArgs, jFn, jArgs):
    wFile = open(fName,'w')
    assert(len(solutions['energies'])==len(solutions['num_occurrences']))
    assert(len(solutions['energies'])==len(solutions['solutions']))
    wFile.write("p solutions 0 "+str(len(solutions['energies']))+" "+str(len(solutions['solutions'][0]))+" "+str(sum([int(it) for it in solutions['num_occurrences']]))+"\n")
    wFile.write("c Legend for line above:\n")
    wFile.write("c   <program marker> <type> <subtype> <nSolutions> <lenSolutions> <nTotalOccurrences>\n")
    wFile.write("c   solutions file written by qforge.py \n")
    wFile.write("c Legend for non-comment lines below:\n")
    wFile.write("c   <nOccurrences> <energy> <nOnes> <solution>\n")
    for i in range(len(solutions['energies'])):
        wFile.write(str(solutions['num_occurrences'][i])+" "+str(solutions['energies'][i])+" "+str(sum([1 if it==1 else 0 for it in solutions['solutions'][i]]))+" "+"".join([str(it)+" " if it in [-1,1] else "_ " for it in solutions['solutions'][i]])+"\n")
    
    #ToDo:  write timing info?
    wFile.flush()
    wFile.close()
    return

def readSolutionFile(args,fName):
    inFile = open(fName,'r')
    # eliminate any initial comments
    line = inFile.next()
    while line[0]=='c':
        line = inFile.next()
    # header line should be next
    hdr = [item.strip() for item in line.split() if len(item)>0]
    if hdr[0]=='p':
        if hdr[1]=='solutions':
            if int(hdr[2])!=0:
                print "unsupported target type '"+str(hdr[1])+"', exiting..."
                sys.exit(-1)
            nSolutions = int(hdr[3])
            lenSolutions = int(hdr[4])
        else:
            print "unsupported file type '"+str(hdr[1])+"', exiting..."
            sys.exit(-1)
    nOccurs = []; energies = []; solns = []; 
    els = 0
    # process the rest of the file.  If annotStr specified, digest those annotations properly
    #  and attach to the following line.
    for line in inFile:
        adj = [item.strip() for item in line.split() if len(item)>0]
        if adj[0].startswith('c'):
            continue
        els += 1
        if len(adj)!=lenSolutions+3:
            print "warning: length of solution "+str(els)+" ("+str(len(adj)-2)+") does not match solution-length declared in header ("+str(lenSolutions)+")"
        nOccurs.append(adj[0])
        energies.append(adj[1])
        #NOTE: skipping element 2, which is the #ones in the solution, since easily recalculated
        solns.append(adj[3:])
        if args.verbose>1: print "els="+str(els)+" "+adj[0]+" "+adj[1]

    # do some sanity-checking on file contents
    if args.verbose>0:
        print "done reading solutions file, #elements="+str(nSolutions)
    if els==0:
        return None
    ret = {'nOccurrences':nOccurs, 'energies':energies, 'solutions':solns}
    return ret



# calculates the energy of a solution with a set of h/Js
def calcEnergy(qubo,soln):
    h = {key[0]:value for (key,value) in sorted(qubo.items()) if key[0]==key[1]}
    J = {key:value for (key,value) in qubo.items() if key[0]!=key[1]}
    e = 0.0
    for (qb1,val) in h.items():
        e += h[qb1]*soln[qb1]
        for (qb2,val) in h.items():
            if qb2 < qb1:
                continue
            if (qb1,qb2) in J:
                e += J[qb1,qb2]*soln[qb1]*soln[qb2]
    return e

ATOMIC_SIZE = 8

# calculate the recursive-bisection signature of the QUBO
def calcRBC(X,Y,args=None):
    nVars = max(max(X),max(Y))+1
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    if nLevels==0:
        return [0]
    signature = [0]*(nLevels)
    newRanges = [(0,nVars/2, nVars/2+1, nVars)]
    ranges = newRanges
    for lvl in range(nLevels-1):
        ranges = newRanges
        # this nested counting loop assumes an upper-triangular matrix
        if args and args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
        for item in ranges:
            for i in range(len(X)):
                if item[0]<=X[i] and X[i]<=item[1] and item[2]<=Y[i] and Y[i]<=item[3]:
                    if args and args.verbose>2: print "cross-bisection: X["+str(i)+"]= "+str(X[i])+", Y[i]= "+str(Y[i])
                    signature[lvl] += 1
        lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
        uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
        newRanges = sorted(lowers+uppers)
  # lastly, count the diagonal-or-above edges that do not cross any bisection
    lvl = nLevels-1
    ranges = newRanges
    if args and args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
    for item in ranges:
        incr = ((item[3]-item[0]+1)*(item[3]-item[0]+2)) / 2
        if args and args.verbose>1: print "  item="+str(item)+", incr="+str(incr)
        for i in range(len(X)):
            if   item[0]<=X[i] and X[i]<=item[3] and item[0]<=Y[i] and Y[i]<=item[3] and X[i]<=Y[i]:
                if args and args.verbose>2: print "within bisection: X["+str(i)+"]= "+str(X[i])+", Y[i]= "+str(Y[i])
                signature[lvl] += 1
        if args and args.verbose>2:  print "signature[lvl]="+str(signature[lvl])+" signComplete[lvl]="+str(signComplete[lvl])
    return signature 

# create a look-up table that, for a given problem-size and vertices i,j, has the level and 
#   index-within-level
def calcRbcLut(nVars):
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    newRanges = [(0,(nVars-1)/2, (nVars-1)/2+1, nVars-1)]
    LUT = {}
    ranges = newRanges
    for lvl in range(nLevels-1):
        ranges = newRanges
        # this nested loop assumes an upper-triangular matrix
        #if args.verbose>0: print "lvl="+str(lvl)+", ranges="+str(ranges)
        for (ndx,item) in zip(range(len(ranges)),ranges):
            for i in range(item[0],item[1]+1):
                for j in range(item[2],item[3]+1):
                    LUT[i,j] = (lvl,ndx)
        lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
        uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
        newRanges = sorted(lowers+uppers)
  # lastly, count the diagonal-or-above edges that do not cross any bisection
    lvl = nLevels-1
    ranges = newRanges
    #if args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
    for (ndx,item) in zip(range(len(ranges)),ranges):
        for i in range(item[0],item[3]+1):
            for j in range(item[0],item[3]+1):
                LUT[i,j] = (lvl,ndx)
    return LUT 

# calculate the bounds of each bisection at each level for a given problem size
def calcRbcBoundsByLevel(nVars):
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    newRanges = [(0,(nVars-1)/2, (nVars-1)/2+1, nVars-1)]
    boundsByLevel = [[] for _ in range(nLevels)]
    ranges = newRanges
    for lvl in range(nLevels-1):
        ranges = newRanges
        boundsByLevel[lvl] = ranges
        # this nested loop assumes an upper-triangular matrix
        lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
        uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
        newRanges = sorted(lowers+uppers)
  # lastly, count the diagonal-or-above edges that do not cross any bisection
    lvl = nLevels-1
    ranges = newRanges
    boundsByLevel[lvl] = ranges
    return boundsByLevel 

# calculate the recursive-bisection signature of the complete graph of a given size
def calcRBCComplete(nVars,args):
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    if nLevels==0:
        return [0]
    signComplete = [0]*(nLevels)
    newRanges = [(0,nVars/2, nVars/2+1, nVars)]
    ranges = newRanges
    for lvl in range(nLevels-1):
        ranges = newRanges
        # this nested counting loop assumes an upper-triangular matrix
        if args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
        for item in ranges:
            signComplete[lvl] += (item[1]-item[0]+1)*(item[3]-item[2]+1)
        lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
        uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
        newRanges = sorted(lowers+uppers)
  # lastly, count the diagonal-or-above edges that do not cross any bisection
    lvl = nLevels-1
    ranges = newRanges
    if args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
    for item in ranges:
        incr = ((item[3]-item[0]+1)*(item[3]-item[0]+2)) / 2
        if args.verbose>1: print "  item="+str(item)+", incr="+str(incr)
        signComplete[lvl] += incr
        if args.verbose>2:  print "signature[lvl]="+str(signature[lvl])+" signComplete[lvl]="+str(signComplete[lvl])
    return signComplete

# find vertices in a graph/QUBO that are not connected
#   Returns: list of unconnected vars if unconnected, else empty list
#ToDo: should take intendedSize as second arg, as can't tell if vertex #N-1 is not used
def findNotConnected(qubo):
    quboSz = 1 + max([it[0][0] if it[0][0] > it[0][1] else it[0][1] for it in qubo.items()]) 
    visited = [False] * quboSz
    startV = qubo.items()[0][0][0]
    frontier = [startV]
    while len(frontier) > 0:
        v = frontier[0]
        frontier = frontier[1:]
        if visited[v]:
            continue
        visited[v] = True
        # find all neighbor vertices of v not yet visited
        possibleV = [it[0][1] if it[0][0]==v else it[0][0] for it in qubo.items() if it[0][0]==v or it[0][1]==v]
        for it in possibleV:
            if not visited[it]:
                frontier.append(it)
    ret = [ndx for (it,ndx) in zip(visited,range(len(visited))) if not it]
    return ret


# determine whether a graph/QUBO is connected
def isConnected(qubo):
    unvisited = findNotConnected(qubo)
    if len(unvisited) == 0:
        ret = True
    else:
        print "Unvisited vertices = "+str(unvisited)
        ret = False
    return ret

# convert the QUBO into edges crossing each bisection 
def old_convert2Levels(args,X,Y):
    if args.profile3:
        import cProfile as cP
        pr = cP.Profile()
        pr.enable()
    nVars = max(max(X),max(Y))
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    # JByLevel is a list of list of lists: the first level is by levels of the RBC, the second is
    #   by #bisections at that level, and the third is the set of edges that cross that bisection
    JByLevel = [[] for _ in range(nLevels)]
    JBoundsByLevel = [[] for _ in range(nLevels)]
    for lvl in range(nLevels):
        JByLevel[lvl] = [[] for _ in range(2**lvl)]
    newRanges = [(0,nVars/2, nVars/2+1, nVars)]
    ranges = newRanges
    for lvl in range(nLevels-1):
        ranges = newRanges
        JBoundsByLevel[lvl] = ranges
        # this nested counting loop assumes an upper-triangular matrix
        if args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
        for (i,item) in zip(range(len(ranges)),ranges):
            #signComplete[lvl] += (item[1]-item[0]+1)*(item[3]-item[2]+1)
            tmpAdd = []
            for j in range(len(X)):
                if item[0]<=X[j] and X[j]<=item[1] and item[2]<=Y[j] and Y[j]<=item[3]:
                    if args.verbose>2: print "cross-bisection: X["+str(j)+"]= "+str(X[j])+", Y[j]= "+str(Y[j])
                    tmpAdd.append((X[j],Y[j]))
            #        JByLevel[lvl][i].append((X[j],Y[j]))
            #tmpAdd = [it2 for it1 in ranges for it2 in zip(X,Y) if it1[0]<=it2[0] <=it1[1] and it1[2]<=it2[1] <=it1[3]]
            JByLevel[lvl][i].extend(tmpAdd)
            
        lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
        uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
        newRanges = sorted(lowers + uppers)
  # lastly, count the diagonal-or-above edges that do not cross any bisection
    lvl = nLevels-1
    ranges = newRanges
    if args.verbose>2: print "lvl="+str(lvl)+", ranges="+str(ranges)
    for (i,item) in zip(range(len(ranges)),ranges):
        #if args.verbose>1: print "  item="+str(item)+", incr="+str(incr)
        tmpAdd = []
        for j in range(len(X)):
            if   item[0]<=X[j]<=item[3] and item[0]<=Y[j]<=item[3] and X[j]<=Y[j]:
                if args.verbose>2: print "within bisection: X["+str(j)+"]= "+str(X[j])+", Y[j]= "+str(Y[j])
                #JByLevel[lvl][i].append((X[j],Y[j]))
                tmpAdd.append((X[j],Y[j]))
        if args.verbose>2:  print "JByLevel["+str(lvl)+"]="+str(JByLevel[lvl])
        JByLevel[lvl][i].extend(tmpAdd)
    if args.profile3:
        pr.disable()
        pr.print_stats(sort='time')
    return (JByLevel,JBoundsByLevel)

# convert the QUBO into edges crossing each bisection 
#   _convert2Levels takes the QUBO edge-dictionary itself
#   convert2Levels takes the dictionary returned by readDIMACSQuboFile, of which the
#     edge-dictionary is one element
#   Note:  because _convert2Levels is intended for use in embedding situations,
#     it ignores self-edges and does not count them in the returned JByLevel list
def _convert2Levels(LUT,qubo):
    #if args.profile3:
        #import cProfile as cP
        #pr = cP.Profile()
        #if pr: pr.enable()
    #quboKeys = qubo['spMat'].keys()
    quboKeys = qubo.keys()
    nVars = 1 + max([it[0] if it[0]>it[1] else it[1] for it in quboKeys])
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    # JByLevel is a list of list of lists: the first level is by levels of the RBC, the second is
    #   by #bisections at that level, and the third is the set of edges that cross that bisection
    JByLevel = [[] for _ in range(nLevels)]
    for lvl in range(nLevels):
        JByLevel[lvl] = [[] for _ in range(2**lvl)]
    for it in quboKeys:
        if it[0]==it[1]:    # ignore self-edges
            continue
        (lvl,ndx) = LUT[it]
        JByLevel[lvl][ndx].append(it)
    #if args.profile3:
        #if pr: pr.disable()
        #pr.print_stats(sort='time')
    return JByLevel

def convert2Levels(LUT,qubo):
        return _convert2Levels(LUT,qubo['spMat'])

# write the embedding out as an ASCII file
def writeEmbedding(args,chains,fName):
    wFile = open(fName,'w')
    wFile.write("c Legend for lines below:\n")
    wFile.write("c   <chain-index> <len(chain)> <qubit numbers in the chain> \n")
    for i in range(len(chains)):
        wFile.write(str(i)+" "+str(len(chains[i]))+" "+' '.join(map(str,chains[i]))+"\n")
    wFile.flush()
    wFile.close()
    return
    
# create sorted lists, for each bisection, of Vs with most cross-bis Es
def createSortedVtxByBis(args,JByLevel,JBoundsByLevel):
    nVars = 1 + max(max([it3[0] for it1 in JByLevel for it2 in it1 for it3 in it2]),max([it3[1] for it1 in JByLevel for it2 in it1 for it3 in it2]))
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    sortedJByBis = [[] for _ in range(nLevels)]
    for lvl in range(nLevels-1):
        sortedJByBis[lvl] = [[] for _ in range(2**lvl)]
    #newRanges = [(0,nVars/2, nVars/2+1, nVars)]
    #ranges = newRanges
    #for lvl in range(nLevels-1):
    #    ranges = newRanges
        # this nested counting loop assumes an upper-triangular matrix
    #    if args.verbose>1: print "lvl="+str(lvl)+", ranges="+str(ranges)
    #    for (i,item) in zip(range(len(ranges)),ranges):
    #        #signComplete[lvl] += (item[1]-item[0]+1)*(item[3]-item[2]+1)
    #        for j in range(len(X)):
    #            if item[0]<=X[j] and X[j]<=item[1] and item[2]<=Y[j] and Y[j]<=item[3]:
    #                if args.verbose>1: print "cross-bisection: X["+str(j)+"]= "+str(X[j])+", Y[j]= "+str(Y[j])
    #                JByLevel[lvl][i].append((X[j],Y[j]))
    #    lowers = [(item[0],(item[1]+item[0])/2,(item[1]+item[0])/2+1,item[1]) for item in ranges]
    #    uppers = [(item[2],(item[3]+item[2])/2,(item[3]+item[2])/2+1,item[3]) for item in ranges]
    #    newRanges = lowers + uppers
  # lastly, count the diagonal-or-above edges that do not cross any bisection
        for i in range(len(JByLevel[lvl])):
            firsts = [it[0] for it in JByLevel[lvl][i]]
            seconds = [it[1] for it in JByLevel[lvl][i]]
            mn0 = min(firsts)
            mx0 = max(firsts)
            mn1 = min(seconds)
            mx1 = max(seconds)
            tmpHist0 = np.histogram(firsts,range(mn0,mx0+2))
            tmpHist1 = np.histogram(seconds,range(mn1,mx1+2))
            sortHist0 = sorted(zip(tmpHist0[0],tmpHist0[1][:-1]),reverse=True)
            sortHist1 = sorted(zip(tmpHist1[0],tmpHist1[1][:-1]),reverse=True)
            sortedJByBis[lvl][i] = (sortHist0,sortHist1)
            #if args.verbose>1: print "foo"
    # handle last (local) level specially
    lvl = nLevels-1
    sortedJByBis[lvl] = [[] for _ in range(2**lvl)]
    for i in range(len(JByLevel[lvl])):
        Vs = [it[0] for it in JByLevel[lvl][i]] + [it[1] for it in JByLevel[lvl][i]]
        mn = min(Vs)
        mx = max(Vs)
        tmpHist = np.histogram(Vs,range(mn,mx+2))
        sortHist = sorted(zip(tmpHist[0],tmpHist[1][:-1]),reverse=True)
        sortedJByBis[lvl][i] = tuple([sortHist])    # odd syntax: to get tuple of a single list
    return sortedJByBis

# create sorted lists, for each bisection, of Vs with most cross-bis Es
def createSortedVtxByCxnty(JByLevel,JBoundsByLevel):
    nVars = 1 + max(max([it3[0] for it1 in JByLevel for it2 in it1 for it3 in it2]),max([it3[1] for it1 in JByLevel for it2 in it1 for it3 in it2]))
    log2NVars = int(math.ceil(math.log(nVars,2)))
    nLevels = int(math.ceil(math.log(float(nVars)/ATOMIC_SIZE,2)))+1
    JByBis = [[] for _ in range(nLevels)]
    for lvl in range(nLevels-1):
        JByBis[lvl] = [[] for _ in range(2**lvl)]
        for i in range(len(JByLevel[lvl])):
            firsts = [it[0] for it in JByLevel[lvl][i]]
            seconds = [it[1] for it in JByLevel[lvl][i]]
            mn0 = JBoundsByLevel[lvl][i][0]
            mx0 = JBoundsByLevel[lvl][i][1]
            mn1 = JBoundsByLevel[lvl][i][2]
            mx1 = JBoundsByLevel[lvl][i][3]
            tmpHist0 = np.histogram(firsts,range(mn0,mx0+2))
            tmpHist1 = np.histogram(seconds,range(mn1,mx1+2))
            JByBis[lvl][i] = ([el for el in tmpHist0[0]],[el for el in tmpHist1[0]])
    # handle last (local) level specially
    lvl = nLevels-1
    JByBis[lvl] = [[] for _ in range(2**lvl)]
    for i in range(len(JByLevel[lvl])):
        Vs = [it[0] for it in JByLevel[lvl][i]] + [it[1] for it in JByLevel[lvl][i]]
        mn = JBoundsByLevel[lvl-1][i/2][i%2*2]
        mx = JBoundsByLevel[lvl-1][i/2][i%2*2+1]
        if len(Vs) > 0:
            tmpHist = np.histogram(Vs,range(mn,mx+2))
            JByBis[lvl][i] = tuple([[el for el in tmpHist[0]]])    # odd syntax: to get tuple of a single list
        else:
            JByBis[lvl][i] = tuple([[0]*(mx-mn+1)])

    # now convert the by-level to a by-vertex by-level degree
    VDegByBis = {}
    for i in range(nVars):
        VDegByBis[i] = []
    for i in range(len(JByBis)):
        vtx = 0
        for j in range(len(JByBis[i])):
            for k in range(len(JByBis[i][j])):
                for it in JByBis[i][j][k]:
                    VDegByBis[vtx].append(it)
                    vtx += 1
        noEntries = [it for it in VDegByBis.items() if len(it[1])<(i+1)]
        for it in noEntries:
            VDegByBis[it[0]].append(0)
    sortedVDeg = sorted(VDegByBis.items(), key=lambda tup:tup[1], reverse=True)
    #return (VDegByBis,sortedVDeg)
    return sortedVDeg

# write the embedding out as an ASCII file
def writeEmbedding(args,chains,fName):
    wFile = open(fName,'w')
    wFile.write("c Legend for lines below:\n")
    wFile.write("c   <chain-index> <len(chain)> <qubit numbers in the chain> \n")
    for i in range(len(chains)):
        wFile.write(str(i)+" "+str(len(chains[i]))+" "+' '.join(map(str,chains[i]))+"\n")
    wFile.flush()
    wFile.close()
    return
    
# read an embedding file as an ASCII file into an list of lists
def readEmbedding(fName):
    ret = []
    rFile = open(fName,'r')
    for line in rFile:
        if line[0]=='c':
            continue
        itemsStr = [item.strip() for item in line.split() if len(item)>0]
        if not unicode(itemsStr[0]).isnumeric():
            raise RuntimeError("first field in line is not numeric")
        items = [int(item) for item in itemsStr]
        if len(items)-2 != items[1]:
            raise RuntimeError("length(chain["+str(items[0])+"]) = "+str(len(items)-2)+", line[2]="+itemsStr[1])
        ret.append(items[2:])
    rFile.close()
    return ret
    
def calcDegree(couplers, chains):
    degree = [0] * len(chains)
    for i in range(len(chains)):
        for j in range(i+1,len(chains)):
            for it1 in chains[i]:
                for it2 in chains[j]:
                    if (it1,it2) in couplers or (it2,it1) in couplers:
                        degree[i] += 1
                        degree[j] += 1
                        break
    return degree

def convertDict2Nx(spMat):
    spMatCSR = {}
    for key,ignore in spMat.items():
        if spMatCSR.get(key[0]) == None:
            spMatCSR[key[0]] = []
        spMatCSR[key[0]].append(key[1])
    G = nx.from_dict_of_lists(spMatCSR)
    for key,value in spMat.items():
        G[key[0]][key[1]]['weight'] = value
    return G


#    minimize the bandwidth of a QUBO
#   currently uses the Cuthill-McKee method
#    ToDo: change to use recursive Kernighan-Lin method
def minBandwidth(qubo):

    # convert spMat (edge-list dict) to ~CSR format as (adj-list dict)
    spMatCSR = {}
    exist = set()
    for key,ignore in qubo.items():
        if spMatCSR.get(key[0]) == None:
            spMatCSR[key[0]] = []
        spMatCSR[key[0]].append(key[1])
        exist.add(key[0])
        exist.add(key[1])
        # ensure that all vars have at least one entry in spMatCSR so 
        #   reverse_cuthill routine will not squeeze them out
        notExist = set(range(max(exist)+1)) - exist
        for it in notExist:
            spMatCSR[it] = [it]     # create a self-edge / diagonal element
    G = nx.from_dict_of_lists(spMatCSR)
    for key,value in qubo.items():
        G[key[0]][key[1]]['weight'] = value
    rcmPerm = list(nxu.reverse_cuthill_mckee_ordering(G))

    origLapl = nx.laplacian_matrix(G)
    origX,origY = np.nonzero(origLapl)

    permLapl = nx.laplacian_matrix(G,nodelist=rcmPerm)
    permX,permY = np.nonzero(permLapl)
    origBW = str((origY-origX).max()+(origX-origY).max()+1)
    minBW = str((permY-permX).max()+(permX-permY).max()+1)
    #if args.verbose>0: 
    print "Original band-width is "+str(origBW)+ "; minimized is "+str(minBW)

    # convert back to edge-list dict
    ret = {}
    for (it0,it1),val in qubo.items():
        ret[min(rcmPerm[it0],rcmPerm[it1]), max(rcmPerm[it0],rcmPerm[it1])] = value
    return ret

#   convert2Csr: convert a QUBO from dictionary form to compressed-sparse-row
#     form to call PyMetis
#   Input:
#   - qubo:
#   Output:
#   - xadj: the indices in adjncy at which each row's elements begin; note
#           this is of length nRows+1
#   - adjncy: the indices of adjacent vertices
#
def convert2Csr(qubo):
    szQubo = size(qubo)
    adj = [[] for _ in range(szQubo)]
    for it in qubo:
        adj[it[0]].append(it[1])
        adj[it[1]].append(it[0])
    xadj = [0]
    adjncy = []
    for it in adj:
        adjncy += sorted(it)
        xadj.append(len(adjncy))
    return (xadj,adjncy)

#   minimizeRbcCuts:  minimizes a QUBO from a recursive-bisection-connectivity
#     point of view;  i.e., the number of edges that cross each bisection.
#     Implemented by calling PyMetis at each level of the graph.
#   Input:
#   - qubo:  the QUBO
#   Output:
#   - minimized QUBO:
#   - dictionary mapping input vertex numbers to output vertex numbers
#   - dictionary mapping output vertex numbers to input vertex numbers
#
def minimizeRbcCuts(qubo,verbose=0):
    import pymetis as mt

    qSz = size(qubo)
    nLvls = nLevels(qSz)
    boundsByLevel = calcRbcBoundsByLevel(qSz)
    vertsByLevel = [[] for _ in range(nLvls)]
    b2sByLevel = [[] for _ in range(nLvls)]
    s2bByLevel = {}
    s2bByLevel[-1] = dict(zip(range(qSz),range(qSz)))
    for i in range(nLvls-1):
        vertsByLevel[i] = [[] for _ in range(2**i)]
        s2bByLevel[i] = {}
        b2sByLevel[i] = {}
        for j in range(2**(i)):
            # designate the vertices that comprise the graph to be partitioned
            rng = [[] for _ in range(boundsByLevel[i][j][3]+1-boundsByLevel[i][j][0])]
            for it in zip(range(boundsByLevel[i][j][3]+1-boundsByLevel[i][j][0]),range(boundsByLevel[i][j][0],boundsByLevel[i][j][3]+1)):
                rng[it[0]] = s2bByLevel[i-1][it[1]]
            (subG,ign1,ign2) = extractSubqubo(qubo,rng)
            xadj,adjncy = convert2Csr(subG)
            parts = mt.part_graph(2, xadj=xadj, adjncy=adjncy)
            part0 = [it[1] for it in zip(parts[1],rng) if it[0]==0]
            part1 = [it[1] for it in zip(parts[1],rng) if it[0]==1]
            if len(part0)-len(part1) > 1:
                # PyMetis does not allow us to specify that we want the partitions to be of equal
                # size.  If they're not equal size (within 1), find the vertices in the long ("tall")
                # section with the fewest cross-bisection edges and move them to the short
                # section
#               if verbose>1: print("minimizeRbcCuts: uneven split, len(part0)="+str(len(part0))+",len(part1)="+str(len(part1)))
                part0Shorter = len(part0) < len(part1)
                (short,tall) = (part0,part1) if part0Shorter else (part1,part0)
                tallECount = [0] * len(tall)
                for ndx,vT in zip(range(len(tall)),tall):
                    for vS in short:
                        if (min(vT,vS),max(vT,vS)) in qubo:
                            tallECount[ndx] += 1
                # tallECount now has the cross-bisection count per-vertex in it.
                tallLeastCxnVs = sorted(zip(tallECount,part1 if part0Shorter else part0))
                nMoveVs = abs(boundsByLevel[i][j][1]-boundsByLevel[i][j][0]+1 - len(part1 if part0Shorter else part0))
                moveVs = [it[1] for it in tallLeastCxnVs[:nMoveVs]]
                for it in moveVs:
                    tall.remove(it)
                    short.append(it)
                (part0,part1) = (tall,short) 
                assert len(part0)>=len(part1),"len(part0)<len(part1)"
            vertsByLevel[i][j] = [(part0,part1)]
            for it in zip(part0+part1,range(boundsByLevel[i][j][0],boundsByLevel[i][j][0]+len(part0)+len(part1))):
                assert it[1] not in s2bByLevel[i], "s2bByLevel["+str(i)+"]["+str(it[1])+"] already set"
                s2bByLevel[i][it[1]] = it[0]
                assert it[0] not in b2sByLevel[i], "b2sByLevel["+str(i)+"]["+str(it[0])+"] already set"
                b2sByLevel[i][it[0]] = it[1]

    outQubo = {}
    outQubo['spMat'] = {}
    for ((v0,v1),val) in qubo.items():
      outRow = min(b2sByLevel[nLvls-2][min(v0,v1)],b2sByLevel[nLvls-2][max(v0,v1)])
      outCol = max(b2sByLevel[nLvls-2][min(v0,v1)],b2sByLevel[nLvls-2][max(v0,v1)])
      assert (outRow,outCol) not in outQubo['spMat'], "("+str(outRow)+","+str(outCol)+") in outQubo['spMat'][nLvls-2]"
      outQubo['spMat'][outRow,outCol] = val
#     if verbose > 2: print("minimizeRbcCuts: just added outQubo['spMat']["+str((outRow,outCol))+"]")

    return (outQubo,b2sByLevel[nLvls-2],s2bByLevel[nLvls-2])
