#!/bin/bash

# Once a parameterized QUBO has been embedded, the libepqmi interface
# allows a C program to execute the QUBO after binding the parameter values.
# This example creates a three-variable QUBO with six parameters and embeds it.
# Then it uses the libeqpmi interface in a C program to read in the QUBO.
# After initialization, the program binds the QUBO parameters and executes
# the resulting QMI three times.

# Exit if not being run via Bash shell:
if [ ! -n "$BASH" ] ; then 
    echo Please run this script $0 with bash; exit 1; 
fi

# Get a connection to the local simulator:
dw set connection laptop

# Get a 128-qubit sampling solver from the simulator:
dw set solver c4-sw_sample

# Make new workspace directory (examples) if it does not exist:
dw mkdir -p examples

# Change to new workspace directory:
dw cd examples

# Initialize the workspace:
dw init

# Create a three variable QUBO which is completely parameterized:
echo "
param: a1
param: b1
param: c1
param: d2
param: e2
param: f2

var: x
var: y
var: z

term: a1 * x
term: b1 * y
term: c1 * z

term: d2 * x * y
term: e2 * x * z
term: f2 * y * z
" > I-qubo.q

# Embed the QUBO:
dw embed I-qubo.q

echo '

#include <stdio.h>
#include <stdlib.h>

#include "epqmi.h"

int main(int argc, char *argv[])
{
  DW_epqmi *epqmi;

  epqmi = DW_epqmi_read(NULL);

  if (epqmi == NULL)
    {
      printf("error: epqmi is NULL\n");
      return 1;
    }

  char **param_names;
  int params;
  DW_epqmi_list_params(epqmi, &param_names, &params);

  int i;
  for (i=0; i<params; ++i) printf("param[%d] is \"%s\"\n", i, param_names[i]);

  float *param_values;
  param_values = (float *) malloc( params * sizeof(float) );
  for (i=0; i<params; ++i)
    {
       float f = (-10 + rand() % 20);
       printf("Setting param \"%s\" to %g\n", param_names[i], f);
       param_values[i] = f;
    }

  DW_epqmi_bind(epqmi, param_values);

  char **var_names;
  int vars;
  DW_epqmi_list_vars(epqmi, &var_names, &vars);

  for (i=0; i<vars; ++i) printf("var[%d] is \"%s\"\n", i, var_names[i]);

  if (DW_epqmi_exec(epqmi, 100))
    {
      fprintf(stderr, "execution error\n");
      return 1;
    }

  int solutions;
  if (DW_epqmi_sols(epqmi, &solutions))
    {
      fprintf(stderr, "solution error\n");
      return 1;
    }
  printf("**************** num_solutions=%d ****************\n", solutions);

  char *var = malloc(vars);

  int s;
  for (s=0; s<solutions; ++s)
    {
      printf("**** solution=%d ****\n", s);
      int valid;
      if (DW_epqmi_sol_vars(epqmi, s, var, &valid))
	{
	  fprintf(stderr, "variable error\n");
	  return 1;
	}
      printf("     valid=%d\n", valid);
      int occurrences;
      DW_epqmi_sol_occurs(epqmi, s, &occurrences);
      float obj;
      DW_epqmi_sol_obj(epqmi, s, &obj);
      printf("     occurs=%d\n", occurrences);
      printf("     objective=%g\n", obj);
      for (i=0; i<vars; ++i) printf("%10s <== %d\n", var_names[i], var[i]);
    }
  
  return 0;
}

' > I-epqmi.c

# Compile I-epqmi.c:
gcc -I $DWAVE_HOME -c I-epqmi.c

# Determine operating system, since the link procedure
# under Windows/MinGW differs from OS X & Linux:
os=$( uname | tr '[:upper:]' '[:lower:]' )

# Link I-epqmi.o with libepqmi.a and SAPI:
if [[ $os == ming* ]]
then
    gcc -o I-epqmi I-epqmi.o $DWAVE_HOME/libepqmi.a $DWAVE_HOME/dwave_sapi.dll
else
    gcc -L $DWAVE_HOME -o I-epqmi I-epqmi.o -lepqmi -ldwave_sapi
fi

# Run I-epqmi three times, each time using different parameter values:
echo "**************** RUN  1 ****************"
./I-epqmi
echo "**************** RUN  2 ****************"
./I-epqmi
echo "**************** RUN  3 ****************"
./I-epqmi
